unit dbCheckFlags;
// dbcheckflags
// dbcheckbox based on specified n-th bit of field's value
// Dadang Sofyan, aa@bitsmart.com
// 20001103

interface

uses WIndows, Messages, Classes, Controls, StdCtrls, db, dbConsts, dbCtrls, SysUtils;

type
  TBitIndex = 0..32-1;

type
  TdbCheckFlags = class(TCustomCheckBox)
  private
    fMaxBits:Integer;
    fCheckBitNo:Integer;
//    fFlagsValue:Integer;
    fDataLink: TFieldDataLink;
    fValueCheck: Integer;
    fValueUncheck: Integer;
    fPaintControl: TPaintControl;
    procedure DataChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetFieldState: TCheckBoxState;
    function GetReadOnly: Boolean;
    function BitNSet(const Bits32:Integer; const BitNo:TBitIndex):Integer;
    function BitNClear(const Bits32:Integer; const BitNo:TBitIndex):Integer;
    function BitNCheck(const Bits32:Integer; const BitNo:TBitIndex):Boolean;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetReadOnly(Value: Boolean);
    procedure SetValueCheck(const Value: Integer);
    procedure SetValueUncheck(const Value: Integer);
    procedure UpdateData(Sender: TObject);
//    function ValueMatch(const ValueList, Value: string): Boolean;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
  protected
    procedure Toggle; override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure WndProc(var Message: TMessage); override;
//    function GetFlagsValue:Integer;
//    procedure SetFlagsValue(Value:Integer); virtual;
    function GetCheckBitNo:Integer; virtual;
    procedure SetCheckBitNo(CheckBitNo:Integer); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Checked;
    property Field: TField read GetField;
    property State;
  published
    property Alignment;
    property AllowGrayed;
    property Caption;
    property Color;
    property Ctl3D;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property ShowHint;
    property TabOrder;
    property TabStop;
//    property ValueChecked: string read fValueCheck write SetValueCheck;
//    property ValueUnchecked: string read fValueUncheck write SetValueUncheck;
    property Visible;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
//    property FieldValue:Integer
//      read GetFlagsValue write SetFlagsValue;
    property CheckBitNo: Integer
      read GetCheckBitNo write SetCheckBitNo;

  end;

procedure Register;

implementation

function tdbCheckFlags.BitNSet(const Bits32:Integer; const BitNo:TBitIndex):Integer; register;
asm//begin Result:= Bits32 or 1 shl BitNo;
  mov EAX, 1
  shl EAX, BitNo
  or EAX, Bits32
end;
function tdbCheckFlags.BitNClear(const Bits32:Integer; const BitNo:TBitIndex):Integer; register;
asm//begin Result:= Bits32 and not (1 shl BitNo);
  mov EAX, 1
  shl EAX, BitNo
  not EAX
  and EAX, Bits32
end;
function tdbCheckFlags.BitNCheck(const Bits32:Integer; const BitNo:TBitIndex):Boolean; register;
asm//begin Result:= Bits32 and (1 shl BitNo) <> 0;
  mov EAX, 1
  shl EAX, BitNo
  and EAX, Bits32
  jz @@end
  mov EAX, 1
@@end:
end;

constructor tdbCheckFlags.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  State := cbUnchecked;
  fCheckBitNo := 0;
  fMaxBits := High(TBitIndex);
  fValueCheck := 1 shl fMaxBits;//STextTrue;
  fValueUncheck := 0;//STextFalse;
  fDataLink := TFieldDataLink.Create;
  fDataLink.Control := Self;
  fDataLink.OnDataChange := DataChange;
  fDataLink.OnUpdateData := UpdateData;
  fPaintControl := TPaintControl.Create(Self, 'BUTTON');
  fPaintControl.Ctl3DButton := True;
end;

destructor tdbCheckFlags.Destroy;
begin
  fPaintControl.Free;
  fDataLink.Free;
  fDataLink := nil;
  inherited Destroy;
end;

procedure tdbCheckFlags.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (fDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

function tdbCheckFlags.GetFieldState: TCheckBoxState;
var  //Text: string;
  I: Integer;
begin
{
  if fDataLink.Field <> nil then
    if fDataLink.Field.IsNull then
      Result := cbGrayed
    else if fDataLink.Field.DataType = ftBoolean then
      if fDataLink.Field.AsBoolean then
        Result := cbChecked
      else
        Result := cbUnchecked
    else
    begin
      Result := cbGrayed;
      Text := fDataLink.Field.Text;
      if ValueMatch(fValueCheck, Text) then Result := cbChecked else
        if ValueMatch(fValueUncheck, Text) then Result := cbUnchecked;
    end
  else
    Result := cbUnchecked;
}
  if not assigned(fDataLink.Field) then
    Result := cbUnchecked
  else
    with fDataLink.Field do
      if IsNUll then Result:= cbGrayed
      else begin
        fMaxBits:= 31;
        case DataType of
          ftString: I:=StrToIntDef(AsString, 0);
          ftSmallInt, ftInteger, ftWord, ftAutoInc: I:= AsInteger;
          ftFloat, ftCurrency, ftBCD: if abs(AsFloat) < MAXINT then I:= trunc(AsFloat)
            else I:= 0;
          ftBoolean: begin
            fMaxBits:=0;
            I:= ord(AsBoolean);
          end;
          else I:= 0;
        end;
          SetValueCheck(I); SetValueUncheck(I);
        if BitNCheck(I, fCheckBitNo) then
          Result:= cbChecked
        else
          Result:= cbUnchecked
      end
end;

procedure tdbCheckFlags.DataChange(Sender: TObject);
begin
  State := GetFieldState;
end;

procedure tdbCheckFlags.UpdateData(Sender: TObject);
//var  Pos: Integer;  S: string;
begin
{
  if State = cbGrayed then
    fDataLink.Field.Clear
  else
    if fDataLink.Field.DataType = ftBoolean then
      fDataLink.Field.AsBoolean := Checked
    else
    begin
      if Checked then S := fValueCheck else S := fValueUncheck;
      Pos := 1;
      fDataLink.Field.Text := ExtractFieldName(S, Pos);
    end;
}
  if assigned(fDataLink.Field) then with fDataLink.Field do
    if State = cbGrayed then Clear
    else begin
      case DataType of
        ftString: if Checked then AsString:= IntToStr(fValueCheck)
          else AsString:= IntToStr(fValueUncheck);
        ftSmallInt, ftInteger, ftWord, ftAutoInc: if Checked then AsInteger:=fValueCheck
          else AsInteger:=fValueUncheck;
        ftFloat, ftCurrency, ftBCD: if Checked then AsInteger:=fValueCheck
          else AsInteger:=fValueUncheck;
        ftBoolean: if Checked then AsBoolean:=Boolean(fValueCheck)
          else AsBoolean:= Boolean(fValueUncheck);
        else;
      end;
    end;
end;

{
function tdbCheckFlags.ValueMatch(const ValueList, Value: string): Boolean;
var  Pos: Integer;
begin
  Result := False;
  Pos := 1;
  while Pos <= Length(ValueList) do
    if AnsiCompareText(ExtractFieldName(ValueList, Pos), Value) = 0 then
    begin
      Result := True;
      Break;
    end;
end;
}

procedure tdbCheckFlags.Toggle;
begin
  if fDataLink.Edit then
  begin
    if fDataLink.Field.IsNull then
      fDataLink.Field.AsInteger:=0;
    inherited Toggle;
    UpdateData(Self);
    fDataLink.Modified;
  end;
end;

function tdbCheckFlags.GetDataSource: TDataSource;
begin
  Result := fDataLink.DataSource;
end;

procedure tdbCheckFlags.SetDataSource(Value: TDataSource);
begin
  fDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

function tdbCheckFlags.GetDataField: string;
begin
  Result := fDataLink.FieldName;
end;

procedure tdbCheckFlags.SetDataField(const Value: string);
begin
  fDataLink.FieldName := Value;
end;

function tdbCheckFlags.GetReadOnly: Boolean;
begin
  Result := fDataLink.ReadOnly;
end;

procedure tdbCheckFlags.SetReadOnly(Value: Boolean);
begin
  fDataLink.ReadOnly := Value;
end;

function tdbCheckFlags.GetField: TField;
begin
  Result := fDataLink.Field;
end;

procedure tdbCheckFlags.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Key of
    #8, ' ':
      fDataLink.Edit;
    #27:
      fDataLink.Reset;
  end;
end;

procedure tdbCheckFlags.SetValueCheck(const Value: Integer);
var I: Integer;
begin
  I:=BitNSet(Value, fCheckBitNo);
  if fValueCheck <> I then begin
    fValueCheck := I;
    DataChange(Self);
  end;
end;

procedure tdbCheckFlags.SetValueUncheck(const Value: Integer);
var I: Integer;
begin
  I:=BitNClear(Value, fCheckBitNo);
  if fValueUncheck <> I then begin
    fValueUncheck := I;
    DataChange(Self);
  end;
end;

procedure tdbCheckFlags.WndProc(var Message: TMessage);
begin
  with Message do
    if (Msg = WM_CREATE) or (Msg = WM_WINDOWPOSCHANGED) or
      (Msg = CM_TEXTCHANGED) or (Msg = CM_FONTCHANGED) then
      fPaintControl.DestroyHandle;
  inherited;
end;

procedure tdbCheckFlags.WMPaint(var Message: TWMPaint);
begin
  if not (csPaintCopy in ControlState) then inherited else
  begin
    SendMessage(fPaintControl.Handle, BM_SETCHECK, Ord(GetFieldState), 0);
    SendMessage(fPaintControl.Handle, WM_PAINT, Message.DC, 0);
  end;
end;

procedure tdbCheckFlags.CMExit(var Message: TCMExit);
begin
  try
    fDataLink.UpdateRecord;
  except
    SetFocus;
    raise;
  end;
  inherited;
end;

procedure tdbCheckFlags.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(fDataLink);
end;

{
function tdbCheckFlags.GetFlagsValue:Integer;
begin
  Result:= fFlagsValue;
end;

procedure tdbCheckFlags.SetFlagsValue(Value: Integer);
begin
  if fFlagsValue <> Value then begin
    fFlagsValue := Value;
    fValueCheck:=BitNSet(Value, fCheckBitNo);
    fValueUnCheck:=BitNClear(Value, fCheckBitNo);
  end;
end;
}

function tdbCheckFlags.GetCheckBitNo:Integer;
begin
  Result:=fCheckBitNo;
  State := GetFieldState;
end;

procedure tdbCheckFlags.SetCheckBitNo(CheckBitNo:Integer);
begin
  if fCheckBitNo <> CheckBitNo then begin
    if CheckBitNo in [0..fMaxBits] then
      fCheckBitNo:= CheckBitNo
    else
      fCheckBitNo:= 0;
    State := GetFieldState;
  end;
end;

procedure Register;
begin
  RegisterComponents('aa', [TdbCheckFlags]);
end;

end.

